part of '../../engine.dart';

/// A way to create and schedule asynchronous events _within_ a game.
///
/// Events are based on _real-life time_, adjusted for game pause durations. For
/// example, time taken for physics or slow motion are _not_ taken into account.
class Scheduler {
  final _scheduled = new Set<Event>.identity();

  @visibleForTesting
  Scheduler();

  /// Cancels and removes any scheduled events.
  @mustCallSuper
  void dispose() {
    _scheduled.clear();
  }

  /// Schedules a [callback] to be invoked after the provided [delay].
  ///
  /// Returns a handle to the [Event] that may be used to cancel it.
  Event schedule(Duration delay, void Function() callback) {
    final event = new Event._(
      delay.inMilliseconds,
      Zone.current.bindCallback(callback),
      this,
    );
    _scheduled.add(event);
    return event;
  }

  /// Elapsed time since the last [update] call.
  num get elapsedTime => _elapsedTime;
  num _elapsedTime = 0;
  num _previousTime;

  /// Executes any scheduled events based on the [currentTime].
  @visibleForTesting
  void update(num currentTime) {
    if (_previousTime != null) {
      _elapsedTime = currentTime - _previousTime;
      for (final event in _scheduled.toList(growable: false)) {
        event._update(_elapsedTime);
      }
    }
    _previousTime = currentTime;
  }
}

/// A single event scheduled by a [Scheduler].
///
/// Events are automatically removed from their parent scheduler once fired.
class Event {
  /// Elapsed time until this event occurs.
  final int _delay;

  /// Callback to be invoked when this event occurs.
  final void Function() _callback;

  /// Parent scheduler, used for cancelling the event.
  final Scheduler _scheduler;

  /// Total elapsed MS.
  ///
  /// When this number is >= [_delay], the event occurs.
  num _totalElapsed = 0;

  Event._(this._delay, this._callback, this._scheduler);

  /// Whether this event is currently active.
  bool get active => _scheduler._scheduled.contains(this);

  /// Cancels the event, returning `true` if this had an effect.
  bool cancel() => _scheduler._scheduled.remove(this);

  /// Updates that [time] has elapsed.
  void _update(num time) {
    if ((_totalElapsed += time) < _delay) {
      return;
    }
    try {
      _callback();
    } finally {
      _scheduler._scheduled.remove(this);
    }
  }
}
