part of '../../geometry.dart';

/// An immutable 2-dimensional line (pair of coordinates).
@Immutable()
class Line {
  /// Start point.
  final Point start;

  /// End point.
  final Point end;

  /// Creates a line from two pairs of x and y coordinates.
  factory Line(int x1, int y1, int x2, int y2) =>
      new Line.fromPoints(new Point(x1, y1), new Point(x2, y2));

  /// Creates a line from existing [Point] objects.
  const Line.fromPoints(this.start, this.end)
      : assert(start != null),
        assert(end != null);

  @override
  bool operator==(Object o) => o is Line && start == o.start && end == o.end;

  @override
  int get hashCode => hash2(start, end);

  /// Angle of the line in radians.
  double get angle => math.atan2(end.y - start.y, end.x - start.x);

  /// Slope of the line (`y/x`).
  double get slope => (end.y - start.y) / (end.x - start.x);

  /// Width of the line.
  int get width => (start.x - end.x).abs();

  /// Height of the line.
  int get height => (start.y - end.y).abs();

  /// Top-most coordinate of this line.
  int get top => math.min(start.y, end.y);

  /// Bottom-most coordinate of this line.
  int get bottom => math.max(start.y, end.y);

  /// Left-most coordinate of this line.
  int get left => math.min(start.x, end.x);

  /// Right-most coordinate of this line.
  int get right => math.max(start.x, end.x);

  @override
  String toString() {
    if (assertionsEnabled) {
      return 'Line {${start.x}, ${start.y}} -> {${end.x}, ${end.y}}';
    }
    return super.toString();
  }
}
